/**************************************************************************
 *									  *
 *    File "neural_slave.c"						  *
 *									  *
 *    Calculation of integral using rectangular segments for P4		  *
 *									  *
 *    Master program in  "neural_master.c"				  *
 *    Process Group File "procgroup" 					  *
 *									  *
 **************************************************************************/
#include  <stdio.h>
#include "p4.h"
#include "cortex.h"
#include "struc.h"
#include <math.h>
   
#define DATA      -1
#define P4_ROW    1
#define P4_COLUMN 4

/* change depending on the number of slaves in file "data_file"  */
#define   HIDDEN_UNIT  16 /* divide into the number of Hidden layer */ 
#define   OUTPUT_UNIT  2 /* divide into the number of Output layer */
 
/*-------------------------------------------------------------------*/
int    A[P4_ROW][P4_COLUMN];

slave()
{
  int    i, j, type, to, from, size, my_id;
  int    k;
  int    l;
  int    m;
  int    n;
  int    s,t;
  int    tuple = 0;
  int    hindex=0, oindex=0;
  float  *incomingf;

  float  wsum_hunit = 0.;
  float  netsum_hunit = 0.;
  float  momentum = 0.9;
  float  learning_rate = 0.1;
  float  wsum_ounit = 0.;
  float  netsum_ounit = 0.;

  /* check if which slave is used */
  my_id = p4_get_my_id();

  /* receive the data from master */
  type = DATA;
  from = 0;  /* master process */
  to = 0;    /* master process */

/* Learning the SCHNN */

/* Compute the output activities of hidden units */
    m=0;
    for ( i = A[0][0]; i < A[0][1]; i++)
      {
        for ( j = 0; j < ITUPLE; j++)
          {
            tuple = HIDDEN->in_tuple [i][j];
            k = tuple / COLUMN;
            l = tuple % COLUMN;
            wsum_hunit += HIDDEN->iweight [i][j] * INPUT->Digit [k][l];
          }
        netsum_hunit = wsum_hunit + HIDDEN->threshold [i];
        HIDDEN->activity[m++] = 1. / ( 1. + exp ( -1. * (double)netsum_hunit ) );
        wsum_hunit = 0.0;
      }

    size = HIDDEN_UNIT * sizeof(float);

    /* The slave is sending the result of activation function for Hidden layer */
    p4_sendr(DATA, to, HIDDEN->activity, size);

  /* receive the data from master */
    size = HUNIT * sizeof(float);
    incomingf = NULL;

    /* The slave is receiving the whole data of activation function for hidden layer */
    p4_recv(&type, &from, &incomingf, &size);
    for ( j = 0 ; j < HUNIT; j++ )
          HIDDEN->activity[j] = incomingf[j];
    p4_msg_free(incomingf);

   /* Compute the output activities of output units */
     n=0;
     for ( j = A[0][2]; j < A[0][3]; j++)
     {
       for ( k = 0; k < HUNIT; k++)
        wsum_ounit += OUTPUT->weight [j][k] * HIDDEN->activity [k];
               
        netsum_ounit = wsum_ounit + OUTPUT->threshold [j];
        OUTPUT->activity [n++] = 1. / ( 1. + exp ( -1. * (double)netsum_ounit ) );
        wsum_ounit = 0.0;
      }

    size = OUTPUT_UNIT * sizeof(float);

    /* The slave is sending the result of activation function for output layer */
    p4_sendr(DATA, to, OUTPUT->activity, size);

  /* receive the data from master */
    size = CATEGORY * sizeof(float);
    incomingf = NULL;

 /* The slave is receiving the whole data of activation function for output layer */
    p4_recv(&type, &from, &incomingf, &size);
    for ( j = 0 ; j < CATEGORY; j++ )
          OUTPUT->activity[j] = incomingf[j];
    p4_msg_free(incomingf);

/* Modifying all the weights and thresholds according to EBP rule */

    /* For output units */
      s = 0;
      for ( i = A[0][2]; i < A[0][3]; i++)
      {
          OUTPUT->error [i] = OUTPUT->desired_activity [i] - 
                              OUTPUT->activity [i];
          OUTPUT->sum_square_error += OUTPUT->error [i] * OUTPUT->error [i];
          OUTPUT->delta [s] = OUTPUT->error [i] * OUTPUT->activity [i] * 
                              ( 1.0 - OUTPUT->activity [i]);
          OUTPUT->delta_bias [i] = learning_rate * 
                                   OUTPUT->delta [s] + momentum * 
                                   OUTPUT->delta_bias [i];
          OUTPUT->threshold [i] += OUTPUT->delta_bias [i];
          for ( k = 0; k < HUNIT; k++)
          {
              OUTPUT->delta_weight [i][k] = learning_rate * 
                      OUTPUT->delta [s] * HIDDEN->activity [k] + 
                      momentum * OUTPUT->delta_weight [i][k];
              OUTPUT->weight [i][k] += OUTPUT->delta_weight [i][k];
          }
          ++s;
      }

    size = OUTPUT_UNIT * sizeof(float);

/* The slave is sending the result of output delta */
    p4_sendr(DATA, to, OUTPUT->delta, size);

    size = HUNIT * sizeof(float);

    /* The slave is sending the result of output weight */
    for (i=A[0][2]; i < A[0][3]; i++) {
    p4_sendr(DATA, to, OUTPUT->weight[i],size); }

  /* receive the data from master */
    size = CATEGORY * sizeof(float);
    incomingf = NULL;

/* The slave is receiving the result of output delta */
    p4_recv(&type, &from, &incomingf, &size);
    for ( j = 0 ; j < CATEGORY; j++ )
          OUTPUT->delta[j] = incomingf[j];
    p4_msg_free(incomingf);

    size = HUNIT * sizeof(float);
    for ( i = 0; i < CATEGORY; i++) {
    incomingf = NULL;

    /* The slave is receiving the whole result of output weight */
    p4_recv(&type, &from, &incomingf, &size);
    for ( j = 0 ; j < HUNIT; j++ )
          OUTPUT->weight[i][j] = incomingf[j];
    p4_msg_free(incomingf); }

    /* For hidden units */
      for ( i = A[0][0]; i < A[0][1]; i++)
      {
          HIDDEN->error_back [i] = 0.;
      }
      for ( i = A[0][0]; i < A[0][1]; i++)
      for ( j = 0; j < CATEGORY; j++)
      {
          HIDDEN->error_back [i] += OUTPUT->delta [j] * 
                                    OUTPUT->weight [j][i];
      }
      for ( i = A[0][0]; i < A[0][1]; i++)
      {
          HIDDEN->delta [i] = HIDDEN->activity [i] * 
                              ( 1.0 - HIDDEN->activity [i] ) * 
                              HIDDEN->error_back [i];
          HIDDEN->delta_bias [i] = learning_rate * HIDDEN->delta [i] + 
                                   momentum * HIDDEN->delta_bias [i];
          HIDDEN->threshold [i] += HIDDEN->delta_bias [i];
          for ( j = 0; j < ITUPLE; j++)
          {
              tuple = HIDDEN->in_tuple [i][j];   
              k = tuple / COLUMN;
              l = tuple % COLUMN;  
              HIDDEN->delta_iweight [i][j] = learning_rate * 
                      HIDDEN->delta [i] * INPUT->Digit [k][l] + 
                      momentum * HIDDEN->delta_iweight [i][j];
              HIDDEN->iweight [i][j] += HIDDEN->delta_iweight [i][j];
          }
      }
}

main (argc, argv)
int     argc;
char    **argv;
{
    register  int   i;               /* Index */
    register  int   j, j1;           /* Index */
    register  int   k;               /* Index */
    register  int   l;               /* Index */
    double     r1 = 0., r2 = 0.;      /* Random number */
    int       time = 0;              /* Processing step */
    int       valid;

    float     wsum_ounit = 0.;
    float     netsum_ounit = 0.;
    float     Hidden_th[40];
    float     Output_th[5];

    int    type, from, size, to;
    int    *incoming;
    int    i1=0,i2=0;
    int    input_target;
    int    input_type;


      p4_initenv(&argc,argv); /* P4 environment is initialized */
      p4_create_procgroup();  /* Process group is created */
    
    /* Creat and Open files */
      stream1 = fopen ( "digit.rf", "r" );
      stream2 = fopen ( "hidden.iwt", "r" );
      stream3 = fopen ( "hidden.ith", "r" );
      stream4 = fopen ( "output.iwt", "r" );
      stream5 = fopen ( "output.ith", "r" );

      stream10 = fopen ( "category.act", "r" );
      stream11 = fopen ( "hidden.tpl","r");
      stream12 = fopen ( "error.0", "w" );
      stream13 = fopen ( "error.3", "w" );
      stream14 = fopen ( "error.6", "w" );
      if ( ( stream15 = fopen ( "error.9", "w" ) ) == NULL )  exit(1);

      INPUT = &cell_i;
      HIDDEN = &cell_h;
      OUTPUT = &cell_o; 

    /* Read the random n-tuples for hidden layer */
      Read_ntuples();

    /* Read the Digit images and categories */
      Read_Digit_image_category();

    /* Initialize the SCHNN */
      Initialize_networks();

  /* receive the data from master */
     type = DATA;
     from = 0;  /* master process */
  
  /*  printf("\n slave %d: ",my_id);  */

  /* receive part of matrix A */
    size = P4_COLUMN * sizeof(int);
    incoming = NULL;
    p4_recv(&type, &from, &incoming, &size);
    for ( i = 0 ; i < P4_ROW; i++)    
    for ( j = 0 ; j < P4_COLUMN ; j++ )
       A[i][j] = incoming[j];
    p4_msg_free(incoming);

    /* STARTING */
      time = 0;
      do {
           /* What category ( 0 - 9 ) ? */
              size = sizeof(int);
              incoming = NULL;
              p4_recv(&type, &from, &incoming, &size);
              input_target = *incoming;
              p4_msg_free(incoming);
 
           /* Mapping Digit's target patterns into output layer */
             Target_mapping(input_target);

           /* Random-orderly training of hand-written digits */
             for ( j1 = 0; j1 < TYPE; j1++)
             {
                   size = sizeof(int);
                   incoming = NULL;
                   p4_recv(&type, &from, &incoming, &size);
                   input_type = *incoming;
                   p4_msg_free(incoming);

                 /* Input token */
                   for ( k = 0; k < ROW; k++)
                   for ( l = 0; l < COLUMN; l++)
                   {
                     INPUT->Digit[k][l] = (float)INPUT->Digit_images[input_target][input_type][k][l];
                   }
                 /* Learning the SCHNN */
                   slave();

             }

           /* Write the average of sum of squared errors and Reset */
             Write_ASSE(input_target);
             
             time++;
      } while ( time <= 0 ); /*Learning the SCHNN(iteration number 4560) */

      for(j = A[0][0]; j < A[0][1]; j++)
         Hidden_th[i1++] = HIDDEN->threshold[j];

      for(j = A[0][2]; j < A[0][3]; j++)
         Output_th[i2++] = OUTPUT->threshold[j];

      to = 0;
      size = HIDDEN_UNIT * sizeof(float);
      p4_sendr(DATA, to, Hidden_th, size);

      size = OUTPUT_UNIT * sizeof(float);
      p4_sendr(DATA, to, Output_th, size);

      size = ITUPLE * sizeof(float);
      for (i=A[0][0]; i < A[0][1]; i++)
        p4_sendr(DATA, to, HIDDEN->iweight[i],size);
      
      p4_wait_for_end();
    /* Close all files and Terminate the process */
      exit(0);

} /* main */
/* =========================================================================
   |                             Functions                                 |
   ========================================================================= */
/* Read the random n-tuples for hidden layer */
void Read_ntuples()
{
    register  int   i;
    register  int   j;

    /* Read n-tuples */
      for ( i = 0; i < HUNIT; i++)
      {
          for ( j = 0; j < ITUPLE; j++)
          {
              fscanf ( stream11, "%4d", &HIDDEN->in_tuple [i][j] );
	  }
      }

    return;
}  /* Read_ntuples */

/* Read the Digit images and categories */
void Read_Digit_image_category()
{
    register  int   i;
    register  int   j;
    register  int   k;
    register  int   l;

    /* Read Digit-images */
      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < TYPE; j++)
      for ( k = 0; k < ROW; k++)
      for ( l = 0; l < COLUMN; l++)
      {
          fscanf ( stream1, "%1d", &INPUT->Digit_images [i][j][k][l] );
      }

    /* Read category patterns */
      for ( i = 0; i < CATEGORY; i++)
      for ( j = 0; j < CATEGORY; j++)
      {
          fscanf ( stream10, "%10f", &INPUT->Digit_category [i][j] );
      }

    return;
} /* Read_Digit_image_category */

/* Initialize the SCHNN */
void Initialize_networks()
{
    register  int   i;
    register  int   j;

    /* For hidden layer */
      for ( i = 0; i < HUNIT; i++)
      {
          fscanf ( stream3, "%10f", &HIDDEN->threshold [i] );
          for ( j = 0; j < ITUPLE; j++)
          {
              fscanf ( stream2, "%10f", &HIDDEN->iweight [i][j] );
          }
      }

    /* For output layer */
      for ( i = 0; i < CATEGORY; i++)
      {
          fscanf ( stream5, "%10f", &OUTPUT->threshold [i] );
          for ( j = 0; j < HUNIT; j++)
          {
              fscanf ( stream4, "%10f", &OUTPUT->weight [i][j] );
          }
      }

      return;
} /* Initialize_networks */

/* Mapping Digit's target pattern into output layer */
void  Target_mapping(i1)
register  int  i1;
{
    register  int   i;               /* Index */

    for ( i = 0; i < CATEGORY; i++ )
    {
        OUTPUT->desired_activity [i] = INPUT->Digit_category [i1][i];
    }

    return;
} /* Target_mapping */

/* Write the average of sum of squared errors and Reset */
void Write_ASSE(i1)
int i1;
{
    OUTPUT->sum_square_error /= (float) TYPE;
    switch ( i1 )
    {
          case 0: fprintf(stream12,"%10.6f\n",OUTPUT->sum_square_error);
                  OUTPUT->sum_square_error = 0.;
                  break;
          case 1: OUTPUT->sum_square_error = 0.;
                  break;
          case 2: OUTPUT->sum_square_error = 0.;
                  break;
          case 3: fprintf(stream13,"%10.6f\n",OUTPUT->sum_square_error);
                  OUTPUT->sum_square_error = 0.;
                  break;
          case 4: OUTPUT->sum_square_error = 0.;
                  break;
          case 5: OUTPUT->sum_square_error = 0.;
                  break;
          case 6: fprintf(stream14,"%10.6f\n",OUTPUT->sum_square_error);
                  OUTPUT->sum_square_error = 0.;
                  break;
          case 7: OUTPUT->sum_square_error = 0.;
                  break;
          case 8: OUTPUT->sum_square_error = 0.;
                  break;
          case 9: fprintf(stream15,"%10.6f\n",OUTPUT->sum_square_error);
                  OUTPUT->sum_square_error = 0.;
                  break;
          default: printf ("ERROR");  exit(1);
    }

    return;
} /* Write_ASSE */

/* END */
/* end of slave */

